import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.HashMap;
import java.util.Map;
import java.util.TreeMap;
import java.util.concurrent.ConcurrentHashMap;
import java.util.Iterator;


public class MockRouter {
    static final int START_TTL = 10;

    final int portNumber;

    private volatile HashMap<Integer, Integer> adjacentDistances = new HashMap<>(); // Port Number, Cost

    private volatile Map<Integer, Integer> neighborLastSpoke = new HashMap<>(); // Port Number, TTL

    int currentSequenceNumber = 0;

    private volatile boolean listenThreadRunning = true;
    private volatile boolean sendThreadRunning = true;

    private volatile Map<Long, String> historyMap = new TreeMap<>();
    long historyStart = 0;

    private volatile Map<Integer, LinkStatePacket> linkStatePackets = new ConcurrentHashMap<>(); // Origin port, LSP

    RoutingTable routingTable;

    public MockRouter(int portNumber, String[] adjacents){
        this.portNumber = portNumber;
        routingTable = new RoutingTable(portNumber);

        // Populate the adjacentDistances map.
        for(String s : adjacents){
            String[] ad = s.split("-", 2);
            this.adjacentDistances.put(Integer.parseInt(ad[0]), Integer.parseInt(ad[1]));
        }

        // Populate the neighborLastSpoke map.
        for(String s : adjacents){
            int p = Integer.parseInt(s.split("-", 2)[0]);
            this.neighborLastSpoke.put(p, START_TTL);
        }

        this.linkStatePackets.put(portNumber, new LinkStatePacket(portNumber, portNumber, currentSequenceNumber++, START_TTL, (Map) adjacentDistances.clone()));

        historyStart = System.nanoTime();

        createListenThread().start();
        createSendThread().start();
    }

    private Thread createListenThread(){
        Thread thread = new Thread(() -> {
            ServerSocket serverSocket = null;
            try {
                serverSocket = new ServerSocket(portNumber);
            } catch (IOException e) {
                e.printStackTrace();
            }
            while (listenThreadRunning) {
                try {
                    // Wait for a client to connect
                    Socket socket = serverSocket.accept();
                    //System.out.printf("%s connected to %d\n",socket.getInetAddress().getHostName(), portNumber);

                    // Make the socket io streams easier to work with.
                    BufferedReader reader = new BufferedReader(new InputStreamReader(socket.getInputStream()));
                    PrintWriter writer = new PrintWriter(new BufferedWriter(new OutputStreamWriter(socket.getOutputStream())), true);

                    // Wait for a line from the client.
                    String line = reader.readLine();
                    //System.out.printf("%d Received: %s\n", portNumber, line);

                    if(line == null || line.length() == 0) continue;
                    if(line.charAt(0) == 'l'){
                        // Store the link state packet text in the history map using the current time.
                        historyMap.put(System.nanoTime() - historyStart, line.substring(2));

                        // If we have never seen this origin before then add it to lsps. If we have seen it, replace it if its newer information.
                        LinkStatePacket lsp = new LinkStatePacket(line);
                        if(!linkStatePackets.containsKey(lsp.getOriginPort())){
                            linkStatePackets.put(lsp.getOriginPort(), lsp);
                        }else{
                            if(lsp.getSequenceNumber() >= linkStatePackets.get(lsp.getOriginPort()).getSequenceNumber()){
                                linkStatePackets.replace(lsp.getOriginPort(), lsp);
                            }
                        }

                        // Reset the neighbor last spoke timer.
                        neighborLastSpoke.replace(lsp.getSenderPort(), START_TTL);

                        // Acknowledge the packet.
                        writer.printf("ACK\n");
                    }else if(line.charAt(0) == 'h'){
                        StringBuilder sb = new StringBuilder();
                        historyMap.forEach((key, value) -> sb.append(String.format("%d: %s\n", key/1000000000, value)));
                        sb.append("TABLE\n");
                        sb.append(routingTable.toString());
                        writer.print(sb);
                        writer.flush();
                    }else if (line.charAt(0) == 's'){
                        // Set flags to false so that both threads will terminate when this session is done.
                        listenThreadRunning = false;
                        sendThreadRunning = false;
                        writer.printf("STOPPING\n");
                    }

                    // Close everything.
                    reader.close();
                    writer.close();
                    socket.close();
                    //System.out.printf("Closed %d\n", portNumber);
                } catch (Exception e) {
                    e.printStackTrace();
                }
            }

            // finally, close the serverSocket
            try {
                serverSocket.close();
            } catch (IOException e) {
                e.printStackTrace();
            }
        });
        return thread;
    }

    private Thread createSendThread(){
        Thread thread = new Thread(() -> {
            while (sendThreadRunning) {
                // Sleep for a random amount of time between 3 and 4 seconds
                try {
                    Thread.sleep((int) (Math.random() * 1000) + 3000);
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }

                // Send all the lsps you have to your neighbors.
                for (Map.Entry<Integer, LinkStatePacket> l : linkStatePackets.entrySet()) {
                    for(Map.Entry<Integer, Integer> n : adjacentDistances.entrySet()){
                        LinkStatePacket lspv = l.getValue();
                        if(lspv.getSenderPort() == n.getKey()) continue; // Don't send lsp from where it came.
                        sendMessage(n.getKey(), lspv.toSenderString(this.portNumber));
                        //System.out.printf("%d wants to talk to %d\n", portNumber, l.getValue().getSenderPort());
                    }
                }

                // Decrement lsp TTLs and throw away expired lsps.
                Iterator<Map.Entry<Integer, LinkStatePacket>> iter = linkStatePackets.entrySet().iterator();
                while(iter.hasNext()){
                    Map.Entry<Integer, LinkStatePacket> l = iter.next();
                    l.getValue().decrementTTL();
                    if(l.getValue().getTimeToLive() <= 0) iter.remove();
                }

                // Decrement neighbor TTLs and throw away neighbors in adj list that have not been heard from.
                for(Map.Entry<Integer, Integer> n : neighborLastSpoke.entrySet()){
                    n.setValue(n.getValue() - 1);
                    if(n.getValue() <= 0) adjacentDistances.remove(n.getKey());
                }

                // Update routers own lsp with list of neighbors and new sequence number.
                this.linkStatePackets.replace(portNumber, new LinkStatePacket(portNumber, portNumber, currentSequenceNumber++, START_TTL, (Map<Integer, Integer>) adjacentDistances.clone()));

                //create routing table using dijkstras and use the information to update this routers routing table.
                routingTable.updateRoutingTable(linkStatePackets);
            }
        });
        return thread;
    }

    private String sendMessage(int port, String message){
        try {
            Socket socket = new Socket("localhost", port);

            BufferedReader reader = new BufferedReader(new InputStreamReader(socket.getInputStream()));
            PrintWriter writer = new PrintWriter(new BufferedWriter(new OutputStreamWriter(socket.getOutputStream())), true);

            writer.println(message);
            String line, accumulate = "";
            while ((line = reader.readLine()) != null) {
                accumulate += line;
            }

            reader.close();
            writer.close();
            socket.close();

            return accumulate;
        } catch (IOException e) {
            // Do nothing, the router has been stopped.
        }

        return null;
    }

    @Override
    public String toString(){
        String s = String.format("%d", portNumber);
        for(Map.Entry<Integer, Integer> e : adjacentDistances.entrySet()){
            s += " " + e.getKey() + "-" + e.getValue();
        }
        return s;
    }
}
