import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

class RoutingTable {
    private static boolean DEBUG = false;

    private ArrayList<EntryTuple> tentative = new ArrayList<>();
    private ArrayList<EntryTuple> confirmed = new ArrayList<>();
    private HashMap<Integer, Integer> previous = new HashMap<>();

    private int startNode = 0;

    public RoutingTable(int startNode){
        this.startNode = startNode;
    }

    public void updateRoutingTable(Map<Integer, LinkStatePacket> linkStatePackets) {
        //First clear out the past information
        tentative.clear();
        confirmed.clear();
        previous.clear();

        //Initialize the Confirmed list with an entry for myself; this entry has a cost of 0.
        confirmed.add(new EntryTuple(startNode, 0, startNode));

        //For the node just added to the Confirmed list in the previous step, call it node Next and select its LSP.
        int nextNode = startNode;
        LinkStatePacket nextLSP = linkStatePackets.get(nextNode);
        previous.put(nextNode, null);

        while(true) {
            if(DEBUG) {
                System.out.println("Confirmed: " + confirmed);
                System.out.println("Tentative: " + tentative);
            }

            /*For each neighbor (Neighbor) of Next, calculate the cost (Cost) to reach this Neighbor as the sum of the cost
              from myself to Next and from Next to Neighbor.*/
            if(nextLSP == null) break;
            for (Map.Entry<Integer, Integer> n : nextLSP.getAdjacentDistances().entrySet()) {
                int neighborID = n.getKey(), neighborCost = n.getValue();
                //cost = myself to next + next to neighbor
                int cost = findEntry(nextNode, confirmed).cost + neighborCost;

                // If Neighbor is currently on neither the Confirmed nor the Tentative list, then add (Neighbor, Cost, NextHop) to the Tentative list, where NextHop is the direction I go to reach Next.
                if (findEntry(neighborID, confirmed) == null && findEntry(neighborID, tentative) == null) {
                    tentative.add(new EntryTuple(neighborID, cost, nextNode));
                }

                // If Neighbor is currently on the Tentative list, and the Cost is less than the currently listed cost for Neighbor, then replace the current entry with (Neighbor, Cost, NextHop), where NextHop is the direction I go to reach Next.
                EntryTuple et = findEntry(neighborID, tentative);
                if (et != null && cost < et.cost) {
                    et.cost = cost;
                    et.nextHop = nextNode;
                }
            }

            // If the Tentative list is empty, stop.
            if(tentative.isEmpty()) break;

            // Otherwise, pick the entry from the Tentative list with the lowest cost, move it to the Confirmed list.
            EntryTuple cheapestEntry = getCheapestEntry(tentative);
            confirmed.add(cheapestEntry);
            tentative.remove(cheapestEntry);

            // For the node just added to the Confirmed list in the previous step, call it node Next and select its LSP.
            previous.put(cheapestEntry.destination, nextNode);

            nextNode = cheapestEntry.destination;
            nextLSP = linkStatePackets.get(nextNode);
        }

        if(DEBUG) {
            System.out.println("Final:");
            System.out.println(confirmed);
            System.out.println(previous);
            System.out.println(getNextHop(1, previous));
            System.out.println(getNextHop(2, previous));
            System.out.println(getNextHop(3, previous));
            System.out.println(getNextHop(4, previous));
        }

    }

    @Override
    public String toString(){
        String s = "destination  cost        nextHop\n-------------------------------------\n";
        for(EntryTuple e : confirmed){
            e.nextHop = getNextHop(e.destination, previous);
            s += e + "\n";
        }
        s += "-------------------------------------\n";
        return s;
    }

    private int getNextHop(int node, HashMap<Integer, Integer> previous){
        Integer next = node;
        Integer last = next;
        while(previous.get(next) != null){
            last = next;
            next = previous.get(next);
        }
        return last;
    }

    private EntryTuple findEntry(int dest, ArrayList<EntryTuple> list){
        for(EntryTuple e : list){
            if(e.destination == dest) return e;
        }
        return null;
    }

    private EntryTuple getCheapestEntry(ArrayList<EntryTuple> list){
        EntryTuple cheapest = list.get(0);
        for(EntryTuple e : list){
            if (e.cost < cheapest.cost) cheapest = e;
        }
        return cheapest;
    }

}

class EntryTuple{
    public int destination, cost, nextHop;
    public EntryTuple(int destination, int cost, int nextHop){
        this.destination = destination;
        this.cost = cost;
        this.nextHop = nextHop;
    }

    @Override
    public String toString(){
        return String.format("|%-11d|%-11d|%-11d|", destination, cost, nextHop);
    }
}
